import logging

import configargparse
from concurrent_log_handler.queue import setup_logging_queues
from prometheus_client import Info, start_http_server

from ._version import __version__
from .authenticator import AADAuthenticatorFlows
from .duo_authenticator import DuoAuthenticator


def main():
    parser = configargparse.ArgParser(
        default_config_files=[
            "/etc/openvpn-auth-azure-ad/config.conf",
            "~/.openvpn-auth-azure-ad",
        ]
    )

    parser.add_argument(
        "-c",
        "--config",
        is_config_file=True,
        help="path of config file",
        env_var="AAD_CONFIG_PATH",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-t",
        "--threads",
        default=10,
        env_var="AAD_THREAD_COUNT",
        help="Amount of threads to handle authentication",
        type=int,
    )

    parser_duo = parser.add_argument_group("duo")
    parser_duo.add_argument(
        "--ikey",
        help="duo integration key",
        env_var="DUO_INTEGRATION_KEY",
    )
    parser_duo.add_argument(
        "--skey",
        help="duo secret key",
        env_var="DUO_SECRET_KEY",
    )
    parser_duo.add_argument(
        "--api-host",
        help="duo api host",
        env_var="DUO_API_HOST",
    )

    parser_authentication = parser.add_argument_group("OpenVPN User Authentication")
    parser_authentication.add_argument(
        "-a",
        "--authenticators",
        default=AADAuthenticatorFlows.DEVICE_TOKEN,
        help="Enable authenticators. Multiple authenticators can be separated with comma",
        env_var="AAD_AUTHENTICATORS",
    )
    parser_authentication.add_argument(
        "--auth-token",
        action="store_true",
        help="Use auth token to re-authenticate clients",
        env_var="AAD_AUTH_TOKEN",
    )
    parser_authentication.add_argument(
        "--auth-token-livetime",
        type=int,
        default=86400,
        help="Livetime of auth tokens in seconds",
        env_var="AAD_AUTH_TOKEN_LIFETIME",
    )
    parser_authentication.add_argument(
        "--remember-user",
        action="store_true",
        help="If user authenticated once, the users refresh token is used to reauthenticate silently if possible.",
        env_var="AAD_REMEMBER_USER",
    )
    parser_authentication.add_argument(
        "--verify-common-name",
        action="store_true",
        help="Check if common_name matches Azure AD UPN",
        env_var="AAD_VERIFY_COMMON_NAME",
    )

    parser_openvpn = parser.add_argument_group("OpenVPN Management Interface settings")
    parser_openvpn.add_argument(
        "-H",
        "--ovpn-host",
        help="Host of OpenVPN management interface.",
        env_var="AAD_OVPN_HOST",
    )
    parser_openvpn.add_argument(
        "-P",
        "--ovpn-port",
        help="Port of OpenVPN management interface.",
        env_var="AAD_OVPN_PORT",
        type=int,
    )
    parser_openvpn.add_argument(
        "-s",
        "--ovpn-socket",
        help="Path of socket or OpenVPN management interface.",
        env_var="AAD_OVPN_SOCKET_PATH",
    )
    parser_openvpn.add_argument(
        "-p",
        "--ovpn-password",
        help="Password for OpenVPN management interface.",
        env_var="AAD_OVPN_PASSWORD",
    )

    parser_prometheus = parser.add_argument_group("Prometheus settings")
    parser_prometheus.add_argument(
        "--prometheus",
        action="store_true",
        env_var="AAD_PROMETHEUS_ENABLED",
        help="Enable prometheus statistics",
    )
    parser_prometheus.add_argument(
        "--prometheus-listen-addr",
        env_var="AAD_PROMETHEUS_LISTEN_HOST",
        default="",
        help="prometheus listen addr",
    )
    parser_prometheus.add_argument(
        "--prometheus-listen-port",
        type=int,
        env_var="AAD_PROMETHEUS_PORT",
        help=" prometheus statistics",
        default=9723,
    )
    parser_prometheus.add_argument(
        "--log-level",
        default=logging.INFO,
        type=lambda x: getattr(logging, x),
        env_var="AAD_LOG_LEVEL",
        help="Configure the logging level.",
    )

    options = parser.parse_args()

    # convert all configured loggers to use a background thread
    setup_logging_queues()

    logging.basicConfig(
        level=options.log_level, format="%(asctime)s %(levelname)s %(message)s"
    )

    if options.prometheus:
        start_http_server(
            options.prometheus_listen_port, options.prometheus_listen_addr
        )
        i = Info("openvpn_auth_azure_ad_version", "info of openvpn-auth-azure-ad")
        i.info({"version": __version__})

    authenticator = DuoAuthenticator(
        ikey=options.ikey,
        skey=options.skey,
        api_host=options.api_host,
        threads=options.threads,
        host=options.ovpn_host,
        port=options.ovpn_port,
        unix_socket=options.ovpn_socket,
        password=options.ovpn_password,
    )
    authenticator.run()



import json
import logging
import time
from typing import Dict, Optional

from cacheout import CacheManager
from msal import PublicClientApplication
from prometheus_client import Counter
from . import util
from ._version import __version__
from .openvpn import ManagementInterface
from .util import errors
from .util.thread_pool import ThreadPoolExecutorStackTraced

openvpn_auth_azure_ad_events = Counter(
    "openvpn_auth_azure_ad_events", "track events", ["event"]
)
openvpn_auth_azure_ad_auth_total = Counter(
    "openvpn_auth_azure_ad_auth_total", "auth total", ["flow"]
)
openvpn_auth_azure_ad_auth_succeeded = Counter(
    "openvpn_auth_azure_ad_auth_succeeded", "auth succeeded", ["flow"]
)
openvpn_auth_azure_ad_auth_failures = Counter(
    "openvpn_auth_azure_ad_auth_failures", "auth failures", ["flow"]
)


log = logging.getLogger(__name__)


class AADAuthenticatorFlows:
    USER_PASSWORD = "username_password"
    DEVICE_TOKEN = "device_token"
    AUTH_TOKEN = "auth_token"


class AADAuthenticator(object):
    token_scopes = ["User.ReadBasic.All"]

    def __init__(
        self,
        app: PublicClientApplication,
        graph_endpoint: str,
        authenticators: str,
        verify_certificate_claim: bool,
        auth_token: bool,
        auth_token_lifetime: int,
        remember_user: bool,
        threads: int,
        host: str = None,
        port: int = None,
        socket: str = None,
        password: str = None,
    ):
        self._app = app
        self._graph_endpoint = graph_endpoint
        self._authenticators = [s.strip() for s in authenticators.split(",")]
        self._openvpn = ManagementInterface(host, port, socket, password)
        self._openvpn.connect()
        self._states = CacheManager(
            {
                "challenge": {"maxsize": 256, "ttl": 600},
                "authenticated": {"maxsize": 256, "ttl": 0},
                "auth_token": {"maxsize": 256, "ttl": 86400},
            }
        )

        self._verify_certificate_claim = verify_certificate_claim
        self._auth_token_enabled = auth_token
        self._auth_token_lifetime = auth_token_lifetime
        self._remember_user_enabled = remember_user
        self._thread_pool = ThreadPoolExecutorStackTraced(max_workers=threads)

    def run(self) -> None:
        log.info("Running openvpn-auth-azure-ad %s" % __version__)
        try:
            while True:
                message = self._openvpn.receive()
                if not message:
                    log.error("Connection to OpenVPN closed. Reconnecting...")
                    self._openvpn.connect(True)
                    continue

                if message.startswith("ERROR:"):
                    log.error(message)
                    continue

                if message.startswith(">CLIENT:DISCONNECT"):
                    self._thread_pool.submit(self.client_disconnect, message)

                elif message.startswith(">CLIENT:CONNECT"):
                    self._thread_pool.submit(self.client_connect, message)

                elif message.startswith(">CLIENT:REAUTH"):
                    self._thread_pool.submit(self.client_reauth, message)

                self._states["challenge"].delete_expired()
                self._states["auth_token"].delete_expired()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            log.error(str(e), exc_info=e)

    def send_authentication_success(self, client: dict) -> None:
        self.log_info(client, "authentication succeeded")
        if client["auth_token"] is not None:
            self._openvpn.send_command(
                'client-auth %s %s\npush "auth-token %s"\nEND'
                % (client["cid"], client["kid"], client["auth_token"])
            )
        else:
            self._openvpn.send_command(
                "client-auth-nt %s %s" % (client["cid"], client["kid"])
            )

    def send_authentication_challenge(self, client: Dict, client_message: str) -> None:
        self.log_debug(client, "authentication challenge: %s" % client_message)

        client_challenge = util.format_client_challenge(client, client_message)
        self._openvpn.send_command(
            'client-deny %s %s "%s" "%s"'
            % (client["cid"], client["kid"], "client_challenge", client_challenge)
        )

    def send_authentication_error(
        self, client: dict, message: str, client_message: Optional[str]
    ) -> None:
        if client_message is None:
            self._openvpn.send_command(
                'client-deny %s %s "%s"' % (client["cid"], client["kid"], message)
            )
        else:
            client_challenge = util.format_client_challenge(client, client_message)
            self._openvpn.send_command(
                'client-deny %s %s "%s" "%s"'
                % (client["cid"], client["kid"], message, client_challenge)
            )

    def setup_auth_token(self, client) -> None:
        if self._auth_token_enabled:
            client["auth_token"] = util.generated_id()
            self._states["auth_token"].set(
                client["auth_token"], client["state_id"], self._auth_token_lifetime
            )

    def verify_client_certificate(self, client, result) -> bool:
        if not self._verify_certificate_claim:
            return True

        if "common_name" not in client["env"]:
            return False

        if "id_token_claims" not in result:
            self.log_warn(client, "Could not get id_token")
            return False

        if self._verify_certificate_claim not in result["id_token_claims"]:
            self.log_warn(
                client,
                "claim %s does not exist in id_token"
                % (self._verify_certificate_claim,),
            )
            return False

        return (
            client["env"]["common_name"]
            == result["id_token_claims"][self._verify_certificate_claim]
        )

    def handle_reauth(self, client: Dict) -> Optional[dict]:
        if not self._auth_token_enabled:
            return None

        openvpn_auth_azure_ad_auth_total.labels(AADAuthenticatorFlows.AUTH_TOKEN).inc()

        # Get Auth-Token from Request
        client["auth_token"] = util.get_auth_token(client)
        if client["auth_token"] is None:
            return None

        # Receive state_id by auth-token
        client["state_id"] = self._states["auth_token"].get(client["auth_token"])
        if client["state_id"] is None:
            return None

        # Get authenticated state
        authenticated_state = self._states["authenticated"].get(client["state_id"])
        if authenticated_state is None:
            return None

        # check is authenticated state has required objects
        if "result" not in authenticated_state or "client" not in authenticated_state:
            return None

        self.log_info(client, "Authenticate using auth-token flow")

        # recheck if connected client is the same as inside the state.
        if (
            client["auth_token"] != authenticated_state["client"]["auth_token"]
            or client["env"]["common_name"]
            != authenticated_state["client"]["env"]["common_name"]
            or client["env"]["username"]
            != authenticated_state["client"]["env"]["username"]
        ):
            return None

        # check if user is active inside azure ad by using a refresh token
        result = self._app.acquire_token_by_refresh_token(
            authenticated_state["result"]["refresh_token"], scopes=self.token_scopes
        )

        return result

    def handle_response_challenge(self, client: Dict) -> Optional[dict]:
        if not client["env"]["password"].startswith("CRV1::"):
            return None

        # get state id from client
        state_id = util.get_state_id(client)
        if state_id is None:
            return None

        # get state from challenge state and check if exists
        state = self._states["challenge"].get(state_id)
        if state is None:
            return None

        # delete used challenge state
        self._states["challenge"].delete(state_id)

        # Check if required properties exist in state
        if "flow" not in state:
            return None

        self.log_info(client, "Continue to authenticate using device token flow")
        result = self.device_token_finish(state)

        return result

    def client_connect(self, data: str) -> None:
        client = AADAuthenticator.parse_client_data(data)
        self.log_info(client, "Received client connect")
        openvpn_auth_azure_ad_events.labels("connect").inc()
        self.authenticate_client(client)

    def client_disconnect(self, data: str) -> None:
        client = AADAuthenticator.parse_client_data(data)
        self.log_info(client, "Received client disconnect event")
        openvpn_auth_azure_ad_events.labels("disconnect").inc()

    def client_reauth(self, data: str) -> None:
        client = AADAuthenticator.parse_client_data(data)
        self.log_info(client, "Received client re auth event")
        openvpn_auth_azure_ad_events.labels("reauth").inc()
        self.authenticate_client(client)

    def authenticate_client(self, client: dict) -> None:
        result = {}

        if client["reason"] == "reauth":
            # allow clients to bypass azure ad authentication by reauthenticate via auth-token
            result = self.handle_reauth(client)
            if result is not None:
                if util.is_authenticated(result):
                    if self.verify_client_certificate(client, result):
                        self.log_info(
                            client, "client certificate does not match Azure AD user."
                        )
                        self.send_authentication_error(
                            client, "client_certificate_not_matched", None
                        )
                        return None

                    openvpn_auth_azure_ad_auth_succeeded.labels(
                        AADAuthenticatorFlows.AUTH_TOKEN
                    ).inc()
                    self.log_info(client, "auth-token flow succeeded")
                    self.send_authentication_success(client)
                    return

        if AADAuthenticatorFlows.DEVICE_TOKEN in self._authenticators:
            result = self.handle_response_challenge(client)
            if result is not None:
                client["state_id"] = util.get_state_id(client)
                if util.is_authenticated(result):
                    if not self.verify_client_certificate(client, result):
                        self.log_info(
                            client, "client certificate does not match Azure AD user."
                        )
                        self.send_authentication_error(
                            client, "client_certificate_not_matched", None
                        )
                        return None

                    self.setup_auth_token(client)
                    self._states["authenticated"].set(
                        client["state_id"], {"client": client, "result": result}
                    )
                    openvpn_auth_azure_ad_auth_succeeded.labels(
                        AADAuthenticatorFlows.DEVICE_TOKEN
                    ).inc()
                    self.log_info(client, "device token flow succeeded")
                    self.send_authentication_success(client)
                    return
                else:
                    openvpn_auth_azure_ad_auth_failures.labels(
                        AADAuthenticatorFlows.DEVICE_TOKEN
                    ).inc()
                    if 70016 in result.get("error_codes", []):
                        error = "device token flow errored: no user action"
                        self.log_info(client, error)
                        self.send_authentication_error(client, error, None)
                    else:
                        self.log_info(
                            client,
                            "device token flow errored: %s "
                            % util.format_error(result),
                        )
                        self.send_authentication_error(
                            client, util.format_error(result), None
                        )
                    return

        client["state_id"] = util.generated_id()
        if self._remember_user_enabled:
            accounts = self._app.get_accounts()
            if "common_name" not in client["env"] and accounts:
                result = self._app.acquire_token_silent(
                    self.token_scopes, account=client["env"]["common_name"]
                )

            if util.is_authenticated(result):
                self.setup_auth_token(client)
                self._states["authenticated"].set(
                    client["state_id"], {"client": client, "result": result}
                )
                self.send_authentication_success(client)
                return

        if AADAuthenticatorFlows.USER_PASSWORD in self._authenticators:
            self.log_debug(client, "Authenticate using username/password flow")
            openvpn_auth_azure_ad_auth_total.labels(
                AADAuthenticatorFlows.USER_PASSWORD
            ).inc()
            result = self._app.acquire_token_by_username_password(
                client["env"]["username"],
                client["env"]["password"],
                scopes=self.token_scopes,
            )

            if util.is_authenticated(result):
                if not self.verify_client_certificate(client, result):
                    self.log_info(
                        client, "client certificate does not match Azure AD user."
                    )
                    self.send_authentication_error(
                        client, "client_certificate_not_matched", None
                    )
                    return None
            else:
                openvpn_auth_azure_ad_auth_failures.labels(
                    AADAuthenticatorFlows.USER_PASSWORD
                ).inc()
                if 65001 in result.get("error_codes", []):
                    # AAD requires user consent for U/P flow
                    error = (
                        "Get consent first: %s"
                        % self._app.get_authorization_request_url(self.token_scopes)
                    )
                    self.send_authentication_error(client, "consent_needed", error)
                else:
                    error = util.format_error(result)
                    self.send_authentication_error(client, error, None)

                self.log_info(client, "password flow errored: %s" % (error,))
                return

            openvpn_auth_azure_ad_auth_succeeded.labels(
                AADAuthenticatorFlows.USER_PASSWORD
            ).inc()
            self.log_info(client, "password flow succeeded")
            # Do not send successful auth command, because device token can be enabled.

        if AADAuthenticatorFlows.DEVICE_TOKEN in self._authenticators:
            self.log_info(client, "Start to authenticate using device token flow")
            flow = self.device_auth_start()
            message = flow["message"] + " Then press OK here. No input required here."

            self.log_debug(client, "Save state as %s" % client["state_id"])
            self._states["challenge"].set(client["state_id"], {"flow": flow})

            openvpn_auth_azure_ad_auth_total.labels(
                AADAuthenticatorFlows.DEVICE_TOKEN
            ).inc()
            self.send_authentication_challenge(client, message)
            return

        self.setup_auth_token(client)
        self._states["authenticated"].set(
            client["state_id"], {"client": client, "result": result}
        )
        self.send_authentication_success(client)
        return

    def device_auth_start(self) -> dict:
        flow = self._app.initiate_device_flow(scopes=self.token_scopes)

        if "user_code" not in flow:
            raise ValueError(
                "Fail to create device flow. Err: %s" % json.dumps(flow, indent=4)
            )

        return flow

    def device_token_finish(self, state: dict) -> dict:
        # Block authentication at least for 30 seconds
        state["flow"]["expires_in"] = 30
        state["flow"]["expires_at"] = time.time() + state["flow"]["expires_in"]

        return self._app.acquire_token_by_device_flow(state["flow"])

    @staticmethod
    def parse_client_data(data: str) -> dict:
        client = {
            "env": {},
            "reason": None,
            "cid": None,
            "kid": None,
            "state_id": None,
        }

        for line in data.splitlines():
            try:
                if line.startswith(">CLIENT:CONNECT") or line.startswith(
                    ">CLIENT:REAUTH"
                ):
                    client_info = line.split(",")
                    client["reason"] = client_info[0].replace(">CLIENT:", "").lower()
                    client["cid"] = client_info[1]
                    client["kid"] = client_info[2]
                elif line.startswith(">CLIENT:DISCONNECT"):
                    client_info = line.split(",")
                    client["reason"] = client_info[0].replace(">CLIENT:", "").lower()
                    client["cid"] = client_info[1]
                elif line.startswith(">CLIENT:ENV,"):
                    client_env = line.split(",")[1].split("=")
                    client["env"][client_env[0]] = (
                        client_env[1] if len(client_env) == 2 else ""
                    )
                else:
                    raise errors.ParseError("Can't parse line: %s" % (line,))
            except Exception:
                raise errors.ParseError("Can't parse line: %s" % (line,))

        return client

    def log_warn(self, client: dict, message: str) -> None:
        prefix = self.format_log_prefix(client)

        log.warning("[%s]: %s" % (prefix, message))

    def log_info(self, client: dict, message: str) -> None:
        prefix = self.format_log_prefix(client)

        log.info("[%s]: %s" % (prefix, message))

    def log_debug(self, client: dict, message: str) -> None:
        prefix = self.format_log_prefix(client)

        log.debug("[%s]: %s" % (prefix, message))

    def format_log_prefix(self, client: dict) -> str:
        prefix = "cid: %s" % (client["cid"],)
        if "common_name" in client["env"]:
            prefix += " | %s" % client["env"]["common_name"]

        return prefix



import logging
from typing import Dict

from cacheout import CacheManager
from prometheus_client import Counter
from duo_client.auth import Auth
from ._version import __version__
from .openvpn import ManagementInterface
from .util import errors
from .util.thread_pool import ThreadPoolExecutorStackTraced

openvpn_duo_events = Counter(
    "openvpn_duo_events", "track events", ["event"]
)
openvpn_duo_auth_total = Counter(
    "openvpn_duo_auth_total", "auth total", ["flow"]
)
openvpn_duo_auth_succeeded = Counter(
    "openvpn_duo_auth_succeeded", "auth succeeded", ["flow"]
)
openvpn_auth_azure_ad_auth_failures = Counter(
    "openvpn_duo_auth_failures", "auth failures", ["flow"]
)


log = logging.getLogger(__name__)


class DuoAuthenticator(object):
    def __init__(
        self,
        ikey: str,
        skey: str,
        api_host: str,
        threads: int,
        host: str = None,
        port: int = None,
        unix_socket: str = None,
        password: str = None,
    ):
        self.ikey = ikey
        self.skey = skey
        self.api_host = api_host
        self.auth = Auth(ikey, skey, api_host)
        self._openvpn = ManagementInterface(host, port, unix_socket, password)
        self._openvpn.connect()
        self._states = CacheManager(
            {
                'challenge': {'maxsize': 256, 'ttl': 600},
                'authenticated': {'maxsize': 256, 'ttl': 0},
                'auth_token': {'maxsize': 256, 'ttl': 86400},
            }
        )
        self._thread_pool = ThreadPoolExecutorStackTraced(max_workers=threads)

    def run(self) -> None:
        log.info('Running openvpn-auth-duo %s', __version__)
        try:
            while True:
                message = self._openvpn.receive()
                if not message:
                    log.error("Connection to OpenVPN closed. Reconnecting...")
                    self._openvpn.connect(True)
                    continue

                if message.startswith("ERROR:"):
                    log.error(message)
                    continue

                if message.startswith(">CLIENT:DISCONNECT"):
                    self._thread_pool.submit(self.client_disconnect, message)

                elif message.startswith(">CLIENT:CONNECT"):
                    self._thread_pool.submit(self.client_connect, message)

                elif message.startswith(">CLIENT:REAUTH"):
                    self._thread_pool.submit(self.client_reauth, message)

                self._states["challenge"].delete_expired()
                self._states["auth_token"].delete_expired()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            log.exception('exception in main thread: %s', e)

    @classmethod
    def parse_client_data(cls, data: str) -> dict:
        client = {
            'env': {},
            'reason': None,
            'cid': None,
            'kid': None,
            'state_id': None,
        }

        for line in data.splitlines():
            try:
                if line.startswith('>CLIENT:CONNECT') or line.startswith(
                    '>CLIENT:REAUTH'
                ):
                    client_info = line.split(',')
                    client['reason'] = client_info[0].replace('>CLIENT:', '').lower()
                    client['cid'] = client_info[1]
                    client['kid'] = client_info[2]
                elif line.startswith('>CLIENT:DISCONNECT'):
                    client_info = line.split(',')
                    client['reason'] = client_info[0].replace('>CLIENT:', '').lower()
                    client['cid'] = client_info[1]
                elif line.startswith('>CLIENT:ENV,'):
                    client_env = line.split(',')[1].split('=')
                    client['env'][client_env[0]] = (
                        client_env[1] if len(client_env) == 2 else ''
                    )
                else:
                    raise errors.ParseError(f"Can't parse line: {line}")
            except Exception:
                raise errors.ParseError(f"Can't parse line: {line}")

        return client

    def vpn_command(self, message):
        self._openvpn.send_command(message)

    def authenticate_client(self, client: Dict):
        username = client['env']['common_name']
        log.info('username: %s', username)
        duo = self.auth.auth('push', username, device='auto')
        log.info('[duo] %s', duo)
        if duo.get('status') == 'allow':
            self.vpn_command(f"client-auth-nt {client['cid']} {client['kid']}")
        else:
            self.vpn_command(
                f'client-deny {client["cid"]} {client["kid"]} "no_response" '
                '"we did not receive a response from duo"'
            )

    def client_connect(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client connect', client['cid'])
        log.info('[%s] Received client connect', client['env']['common_name'])
        openvpn_duo_events.labels('connect').inc()
        self.authenticate_client(client)

    def client_disconnect(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client disconnect event', client['cid'])
        openvpn_duo_events.labels('disconnect').inc()

    def client_reauth(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client reauth event', client['cid'])
        openvpn_duo_events.labels('reauth').inc()
        self.vpn_command(f"client-auth-nt {client['cid']} {client['kid']}")



from datetime import datetime, timedelta
import logging
from base64 import b64encode, b64decode
from typing import Dict
from hashlib import sha256
from cacheout import CacheManager
import pytz
from prometheus_client import Counter
from duo_client.auth import Auth
from pyotp import TOTP
from mysql.connector import connect
from mysql.connector import Error as MysqlError
from mysql.connector.errors import OperationalError
from ._version import __version__
from .openvpn import ManagementInterface
from .util import errors, b64encode_string, generated_id
from .util.thread_pool import ThreadPoolExecutorStackTraced


openvpn_totp_events = Counter(
    "openvpn_totp_events", "track events", ["event"]
)
openvpn_totp_auth_total = Counter(
    "openvpn_totp_auth_total", "auth total", ["flow"]
)
openvpn_totp_auth_succeeded = Counter(
    "openvpn_totp_auth_succeeded", "auth succeeded", ["flow"]
)
openvpn_totp_auth_failures = Counter(
    "openvpn_totp_auth_failures", "auth failures", ["flow"]
)


log = logging.getLogger(__name__)


class TotpAuthenticator(object):
    def __init__(
        self,
        mysql_host: str,
        mysql_username: str,
        mysql_password: str,
        mysql_database: str,
        threads: int,
        host: str = None,
        port: int = None,
        unix_socket: str = None,
        password: str = None,
        token_expiration = 15,
    ):
        self.mysql_host = mysql_host
        self.mysql_username = mysql_username
        self.mysql_password = mysql_password
        self.mysql_pool_name = 'totp_authenticator'
        self.mysql_pool_size = threads
        self.mysql_database = mysql_database
        self.token_expiration = token_expiration
        if (host and port) or unix_socket:
            self._openvpn = ManagementInterface(host, port, unix_socket, password)
            self._openvpn.connect()
        self._states = CacheManager({
            'challenge': {'maxsize': 256, 'ttl': 600},
            'authenticated': {'maxsize': 256, 'ttl': 0},
            'auth_token': {'maxsize': 256, 'ttl': 86400},
        })
        self._thread_pool = ThreadPoolExecutorStackTraced(max_workers=threads)

    def run(self) -> None:
        log.info('Running openvpn-auth-duo %s', __version__)
        try:
            while True:
                message = self._openvpn.receive()
                if not message:
                    log.error('Connection to OpenVPN closed. Reconnecting...')
                    self._openvpn.connect(True)
                    continue

                if message.startswith('ERROR:'):
                    log.error(message)
                    continue

                if message.startswith('>CLIENT:DISCONNECT'):
                    self._thread_pool.submit(self.client_disconnect, message)

                elif message.startswith('>CLIENT:CONNECT'):
                    self._thread_pool.submit(self.client_connect, message)

                elif message.startswith('>CLIENT:REAUTH'):
                    self._thread_pool.submit(self.client_reauth, message)

                self._states['challenge'].delete_expired()
                self._states['auth_token'].delete_expired()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            log.exception('exception in main thread: %s', e)

    def mysql_connection(self):
        return connect(
            pool_name=self.mysql_pool_name,
            pool_size=self.mysql_pool_size,
            pool_reset_session=False,
            user=self.mysql_username,
            password=self.mysql_password,
            host=self.mysql_host,
            database=self.mysql_database,
            autocommit=True,
            time_zone='+00:00',
        )

    def query(self, sql, params=None):
        try:
            with self.mysql_connection() as conn:
                with conn.cursor() as cursor:
                    cursor.execute(sql, params)
                    data = []
                    if cursor.description:
                        header = [ x[0].lower().strip() for x in cursor.description ]
                        for x in cursor.fetchall():
                            data.append(dict(zip(header, x)))
                    return data
        except OperationalError as ex:
            log.exception('exception in main thread: %s', ex)
            log.exception('ex.errno: %s', ex.errno)
            log.exception('ex.msg: %s', ex.msg)
        except Exception as ex:
            log.info('error: %s.%s', ex.__module__, ex.__class__.__name__)
            log.info('error: %s', ex)

    @classmethod
    def parse_client_data(cls, data: str) -> dict:
        client = {
            'env': {},
            'reason': None,
            'cid': None,
            'kid': None,
            'state_id': None,
        }

        for line in data.splitlines():
            try:
                if line.startswith('>CLIENT:CONNECT') or line.startswith(
                    '>CLIENT:REAUTH'
                ):
                    client_info = line.split(',')
                    client['reason'] = client_info[0].replace('>CLIENT:', '').lower()
                    client['cid'] = client_info[1]
                    client['kid'] = client_info[2]
                elif line.startswith('>CLIENT:DISCONNECT'):
                    client_info = line.split(',')
                    client['reason'] = client_info[0].replace('>CLIENT:', '').lower()
                    client['cid'] = client_info[1]
                elif line.startswith('>CLIENT:ENV,'):
                    client_env = line.split(',')[1].split('=')
                    client['env'][client_env[0]] = (
                        client_env[1] if len(client_env) == 2 else ''
                    )
                else:
                    raise errors.ParseError(f"Can't parse line: {line}")
            except Exception:
                raise errors.ParseError(f"Can't parse line: {line}")

        return client

    def vpn_command(self, message):
        self._openvpn.send_command(message)

    def last_login(self, username, ip):
        results = self.query(
            'select last_sign_in from last_sign_in where email=%s and ip_address=%s'
            ' order by last_sign_in desc',
            [ username, ip ])
        if results:
            result = results[0]
            return result['last_sign_in'].replace(tzinfo=pytz.utc)
        return None

    def authenticated(self, client, last):
        env = client['env']
        username = env['common_name']
        untrusted_ip = env['untrusted_ip']
        self.vpn_command(f"client-auth-nt {client['cid']} {client['kid']}")
        self.save_last_login(username, untrusted_ip, last)

    def authenticate_client(self, client: Dict):
        env = client['env']
        username = env['common_name']
        password = env['password']
        log.info('username: %s', username)
        log.debug('password: %s', password)
        for key, value in env.items():
            log.debug('[env] %s => %s', key, value)
        untrusted_ip = env['untrusted_ip']
        last = self.last_login(username, untrusted_ip)
        delta = timedelta(days=self.token_expiration)
        if last and datetime.now(tz=pytz.utc) - last <= delta:
            # if a user has signed in from this ip within the last 15 days
            # don't request another otp code
            self.vpn_command(f"client-auth-nt {client['cid']} {client['kid']}")
            return
        crv1 = password.startswith('CRV1') or password.startswith('SCRV1')
        totp_in_password = len(password) == 6 and password.isdigit()
        log.info('crv1: %s', crv1)
        log.info('totp_in_password: %s', totp_in_password)
        if crv1 or totp_in_password:
            results = self.query(
                'select * from totp where email=%s',
                [ username ])
            log.info('[results] found %s results', len(results))
            if not results:
                log.info('username not found in totp table, denying login')
                self.vpn_command(
                    f'client-deny {client["cid"]} {client["kid"]} "no_response" '
                    '"user not authorized for logon"'
                )
                return
            result = results[0]
            otp = TOTP(result['secret_key'])
            pieces = password.split(':')
            totp_response = pieces[-1]
            log.info('response: %s', totp_response)
            if pieces[0] == 'SCRV1':
                response_bytes = totp_response.encode('utf-8')
                totp_response = b64decode(response_bytes).decode('utf-8')
                log.info('response: %s', totp_response)
            if otp.verify(totp_response):
                self.authenticated(client, last)
            else:
                self.vpn_command(
                    f'client-deny {client["cid"]} {client["kid"]} "bad_response" '
                    '"incorrect otp"'
                )
            return
        self.send_client_challenge(client, 'Please enter your one-time code')

    def send_client_challenge(self, client: dict, challenge):
        username = client['env']['username']
        username_b64 = b64encode_string(username)
        state_id = sha256(username.encode('utf-8')).digest()
        state_id = b64encode(state_id).decode('utf-8')
        state_id = generated_id()
        challenge = f'CRV1:E,R:{state_id}:{username_b64}:{challenge}'
        self.vpn_command(
            f'client-deny {client["cid"]} {client["kid"]} '
            f'"client_challenge" "{challenge}"')

    def client_connect(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client connect', client['cid'])
        log.info('[%s] Received client connect', client['env']['common_name'])
        openvpn_totp_events.labels('connect').inc()
        self.authenticate_client(client)

    def client_disconnect(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client disconnect event', client['cid'])
        openvpn_totp_events.labels('disconnect').inc()

    def client_reauth(self, data: str) -> None:
        client = self.parse_client_data(data)
        log.info('[%s] Received client reauth event', client['cid'])
        openvpn_totp_events.labels('reauth').inc()
        self.vpn_command(f"client-auth-nt {client['cid']} {client['kid']}")

    def save_last_login(self, username, ip, last):
        if last:
            self.query(
                'update last_sign_in'
                ' set last_sign_in=current_timestamp'
                ' where '
                '   email=%s'
                '   and ip_address=%s', [username, ip, ]
            )
        else:
            self.query(
                'insert into last_sign_in (email, ip_address)'
                ' values (%s, %s)', [username, ip, ]
            )




import contextlib
import logging
import queue
import re
import socket
import threading
import time
from typing import Generator, Optional
from ..util import errors

logger = logging.getLogger(__name__)

FIRST_LINE_REGEX = re.compile(r'^>CLIENT:(?P<event>([^,]+))(.*)$')
LAST_LINE_REGEX = re.compile(r'^(?:>CLIENT:ENV,)?END$')


class SocketType:
    IP = 'ip'
    UNIX_SOCKET = 'socket'


class ManagementInterface(object):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        socket_path: Optional[str] = None,
        password: Optional[str] = None,
    ):
        if (
            (socket_path and host)
            or (socket_path and port)
            or (not socket_path and not host and not port)
        ):
            raise errors.AuthenticatorError(
                'Must specify either socket or host and port'
            )
        if socket_path:
            self._mgmt_socket = socket_path
            self._type = SocketType.UNIX_SOCKET
        else:
            self._mgmt_host = host
            self._mgmt_port = port
            self._type = SocketType.IP
        self._mgmt_password = password
        self._socket = None
        self._release = None

        self._socket_file = None
        self._socket_io_lock = threading.Lock()

        self._listener_thread = None
        self._writer_thread = None

        self._recv_queue = {
            'COMMAND': queue.Queue(),
            'CLIENT': queue.Queue(),
            'INFO': queue.Queue(),
        }
        self._send_queue = queue.Queue()

    @property
    def type(self) -> Optional[str]:
        '''Get SocketType object for this VPN.
        '''
        return self._type

    @property
    def mgmt_address(self) -> str:
        '''Get address of management interface.
        '''
        if self.type == SocketType.IP:
            return f'{self._mgmt_host}:{self._mgmt_port}'
        return str(self._mgmt_socket)

    def connect(self, retry: bool = False) -> Optional[bool]:
        '''Connect to management interface socket.
        '''
        while True:
            try:
                if self.type == SocketType.IP:
                    self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                    self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                    self._socket.connect(
                        ('{}'.format(self._mgmt_host), int(self._mgmt_port))
                    )
                else:
                    self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                    self._socket.connect(self._mgmt_socket)

                resp = self._socket.recv(1024).decode('utf-8')

                self._socket_file = self._socket.makefile('r')
                self._listener_thread = threading.Thread(
                    target=self._socket_listener_thread,
                    daemon=True,
                    name='mgmt-listener',
                )
                self._writer_thread = threading.Thread(
                    target=self._socket_writer_thread, daemon=True, name='mgmt-writer'
                )

                self._listener_thread.start()
                self._writer_thread.start()

                if resp.startswith('ENTER PASSWORD'):
                    resp = self.send_command(self._mgmt_password)
                    if not resp.startswith('SUCCESS: password is correct'):
                        logger.critical('Wrong management interface password.')

                    assert resp.startswith(
                        'SUCCESS: password is correct'
                    ), 'Wrong management interface password.'

                    resp = self._socket_recv('INFO')
                    assert resp.startswith(
                        '>INFO'
                    ), 'Did not get expected response from interface when opening socket.'
                else:
                    assert resp.startswith(
                        '>INFO'
                    ), 'Did not get expected response from interface when opening socket.'

                self._get_version()
                logger.info('Connection to OpenVPN management interfaced established.')

                return True
            except AssertionError as e:
                if not retry:
                    raise e
            except (socket.timeout, socket.error) as e:
                if retry:
                    logger.error(str(e))
                    time.sleep(1)
                else:
                    raise errors.ConnectError(str(e)) from None

    def disconnect(self, _quit=True) -> None:
        '''Disconnect from management interface socket.
        '''
        if self._socket is not None:
            if _quit:
                self._socket_send('quit\n')
            self._socket.close()
            self._socket = None

    @property
    def is_connected(self) -> bool:
        '''Determine if management interface socket is connected or not.
        '''
        return self._socket is not None

    @property
    def release(self) -> str:
        '''OpenVPN release string.
        '''
        if self._release is None:
            self._release = self._get_version()
        return self._release

    @contextlib.contextmanager
    def connection(self) -> Generator:
        '''Create context where management interface socket is open and close when done.
        '''
        self.connect()
        try:
            yield
        finally:
            self.disconnect()

    def _socket_listener_thread(self):
        '''This thread handles the socket's output and handles any events before adding the output to the receive queue.
        '''
        recv_lines = []
        while True:
            if not self.is_connected:
                break

            line = self._socket_file.readline().strip()

            if not line:
                self._recv_queue['CLIENT'].put(None)
                break

            if line.startswith('>INFO:'):
                self._recv_queue['INFO'].put(line + '\n')
                continue

            if (
                line.startswith('SUCCESS:')
                or line.startswith('ERROR:')
                or line.startswith('ENTER PASSWORD')
            ):
                self._recv_queue['COMMAND'].put(line + '\n')
                continue

            if len(recv_lines) == 0:
                self._socket_io_lock.acquire()

            recv_lines.append(line)

            if LAST_LINE_REGEX.match(line):
                self._socket_io_lock.release()
                queue_name = 'CLIENT' if line.startswith('>CLIENT') else 'COMMAND'
                self._recv_queue[queue_name].put('\n'.join(recv_lines))
                recv_lines = []

    def _socket_writer_thread(self):
        while True:
            if not self.is_connected:
                break

            try:
                data = self._send_queue.get()
                self._socket_io_lock.acquire()
                self._socket.send(bytes(data, 'utf-8'))
            finally:
                self._socket_io_lock.release()

    def _socket_send(self, data) -> None:
        '''Convert data to bytes and send to socket.
        '''
        if self._socket is None:
            raise errors.NotConnectedError(
                'You must be connected to the management interface to issue commands.'
            )
        self._send_queue.put(data)

    def _socket_recv(self, queue_name: str) -> str:
        '''Receive bytes from socket and convert to string.
        '''
        if self._socket is None:
            raise errors.NotConnectedError(
                'You must be connected to the management interface to issue commands.'
            )

        return self._recv_queue[queue_name].get()

    def send_command(self, cmd) -> Optional[str]:
        '''Send command to management interface and fetch response.
        '''
        if not self.is_connected:
            raise errors.NotConnectedError(
                'You must be connected to the management interface to issue commands.'
            )
        logger.debug('Sending cmd: %r', cmd.strip())
        self._socket_send(cmd + '\n')
        if cmd.startswith('kill') or cmd.startswith('client-kill'):
            return
        resp = self._socket_recv('COMMAND')
        logger.debug('Cmd response: %r', resp)
        return resp

    def _get_version(self) -> str:
        '''Get OpenVPN version from socket.
        '''
        raw = self.send_command('version')
        for line in raw.splitlines():
            if line.startswith('OpenVPN Version'):
                return line.replace('OpenVPN Version: ', '')
        raise errors.ParseError(
            'Unable to get OpenVPN version, no matches found in socket response.'
        )

    def receive(self) -> str:
        '''Poll for incoming data
        '''
        if not self.is_connected:
            raise errors.NotConnectedError(
                'You must be connected to the management interface to issue commands.'
            )
        logger.debug('Waiting for incoming data')

        return self._socket_recv('CLIENT')

    @staticmethod
    def has_prefix(line) -> bool:
        return (
            line.startswith('>INFO')
            or line.startswith('>CLIENT')
            or line.startswith('>STATE')
        )

    @property
    def socket(self):
        return self._socket



class AuthenticatorError(Exception):
    """Base exception for all other project exceptions."""


class NotConnectedError(AuthenticatorError):
    """Exception raised if not connected to the management interface and a command is called."""


class ConnectError(AuthenticatorError):
    """Exception raised on connection failure."""


class ParseError(AuthenticatorError):
    """Exception raised on parse error."""



import base64
import uuid

from typing import Optional


def b64encode_string(s: str) -> str:
    return base64.b64encode(s.encode("utf-8")).decode("utf-8")


def is_authenticated(result: dict) -> bool:
    return "access_token" in result


def generated_id() -> str:
    return str(uuid.uuid4())


def format_error(result: dict) -> str:
    return result.get("error_description").splitlines()[0].strip()


def get_state_id(client: dict) -> Optional[str]:
    if "password" not in client["env"]:
        return None

    password = client["env"]["password"].split("::")
    if len(password) < 2:
        return None

    return password[1]


def get_auth_token(client: dict) -> Optional[str]:
    if "password" not in client["env"]:
        return None

    return client["env"]["password"]


def format_client_challenge(client: dict, challenge) -> str:
    username_b64 = b64encode_string(client["env"]["username"])
    return f'CRV1:E,R:{client["state_id"]}:{username_b64}:{challenge}'



# -*- coding: utf-8 -*-
import os
import setuptools  # type: ignore

from openvpn_auth_duo._version import __version__

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="openvpn-auth-duo",
    version=__version__,
    entry_points={
        'console_scripts': [
            'openvpn-auth-duo = openvpn_auth_duo.duo_cli:main',
            'openvpn-auth-totp = openvpn_auth_duo.totp_cli:main',
        ],
    },
    license="MIT",
    author="crazy penguins",
    author_email="pshingavi@gmail.com,",
    description="openvpn-auth-duo connects to the openvpn management interface and handle the authentication "
    "against duo.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/crazy-penguins/openvpn-auth-duo",
    download_url="https://github.com/crazy-penguins/openvpn-auth-duo/archive/v%s.tar.gz"
    % __version__,
    packages=setuptools.find_packages(),
    package_data={
        'openvpn_auth_duo': [ '*.sql', ]
    },
    keywords=["OpenVPN", "AzureAD", "authentication", 'duo'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Topic :: System :: Systems Administration :: Authentication/Directory",
    ],
    install_requires=[
        'msal~=1.4',
        'cacheout~=0.11',
        'ConfigArgParse~=1.2',
        'prometheus_client~=0.8',
        'concurrent-log-handler~=0.9',
        'duo-client',
        'mysql-connector-python',
        'pyotp',
        'setuptools',
        'qrcode',
        'pytz',
    ],
    project_urls={
        "Changelog": "https://github.com/crazy-penguins/openvpn-auth-duo/blob/v%s/CHANGELOG.md"
        % __version__,
        "Source": "https://github.com/crazy-penguins/openvpn-auth-duo",
        "Bug Reports": "https://github.com/crazy-penguins/openvpn-auth-duo/issues",
    },
)



import io
import logging
from socket import gethostname

import configargparse
from concurrent_log_handler.queue import setup_logging_queues
from prometheus_client import Info, start_http_server
from pyotp import random_base32
from pyotp import TOTP
from qrcode import QRCode
from ._version import __version__
from .totp_authenticator import TotpAuthenticator


def qr_code(st):
    qr = QRCode()
    qr.add_data(st)
    f = io.StringIO()
    qr.print_ascii(out=f)
    f.seek(0)
    print(f.read())


def main():
    log = logging.getLogger(__name__)
    parser = configargparse.ArgParser(
        default_config_files=[
            '/etc/openvpn-auth-duo/config.conf',
            '~/.openvpn-auth-duo',
        ]
    )

    parser.add_argument(
        '-c',
        '--config',
        is_config_file=True,
        help='path of config file',
        env_var='AAD_CONFIG_PATH',
    )
    parser.add_argument(
        '-V',
        '--version',
        action='version',
        version='%(prog)s {version}'.format(version=__version__),
    )
    parser.add_argument(
        '-t',
        '--threads',
        default=10,
        env_var='AAD_THREAD_COUNT',
        help='Amount of threads to handle authentication',
        type=int,
    )

    parser_duo = parser.add_argument_group('commands')
    parser_duo.add_argument(
        '--enroll',
        help='switch which, if specified, enrolls a user for totp',
        default=False,
        action='store_true',
    )
    parser_duo.add_argument(
        '--reenroll',
        help='switch which, if specified, reenrolls a user for totp and '
             'generates a new totp secret key',
        default=False,
        action='store_true',
    )
    parser_duo.add_argument(
        '--email',
        help='the email address of the person to enroll',
    )
    parser_duo.add_argument(
        '--migrate',
        help='migrates the schema to the latest version',
        default=False,
        action='store_true',
    )

    parser_duo = parser.add_argument_group('mysql')
    parser_duo.add_argument(
        '--mysql-host',
        help='mysql host',
        env_var='OVAD_MYSQL_HOST',
        default='127.0.0.1',
    )
    parser_duo.add_argument(
        '--mysql-username',
        help='username used to login to the mysql host',
        env_var='OVAD_MYSQL_USERNAME',
    )
    parser_duo.add_argument(
        '--mysql-password',
        help='password to login to mysql ',
        env_var='OVAD_MYSQL_PASSWORD',
        default='',
    )
    parser_duo.add_argument(
        '--mysql-database',
        default='totp',
        help='database housing the totp schema',
        env_var='OVAD_MYSQL_DATABASE',
    )

    parser_authentication = parser.add_argument_group('OpenVPN User Authentication')
    parser_authentication.add_argument(
        '--auth-token',
        action='store_true',
        help='Use auth token to re-authenticate clients',
        env_var='AAD_AUTH_TOKEN',
    )
    parser_authentication.add_argument(
        '--token-expiration',
        type=int,
        default=15,
        help='the number of days a user can sign in without the TOTP after '
             'a successful login with totp',
        env_var='TOTP_TOKEN_EXPIRATION',
    )
    parser_authentication.add_argument(
        '--remember-user',
        action='store_true',
        help='If user authenticated once, the users refresh token is '
             'used to reauthenticate silently if possible.',
        env_var='AAD_REMEMBER_USER',
    )
    parser_authentication.add_argument(
        '--verify-common-name',
        action='store_true',
        help='Check if common_name matches Azure AD UPN',
        env_var='AAD_VERIFY_COMMON_NAME',
    )

    parser_openvpn = parser.add_argument_group('OpenVPN Management Interface settings')
    parser_openvpn.add_argument(
        '-H',
        '--ovpn-host',
        help='Host of OpenVPN management interface.',
        env_var='AAD_OVPN_HOST',
    )
    parser_openvpn.add_argument(
        '-P',
        '--ovpn-port',
        help='Port of OpenVPN management interface.',
        env_var='AAD_OVPN_PORT',
        type=int,
    )
    parser_openvpn.add_argument(
        '-s',
        '--ovpn-socket',
        help='Path of socket or OpenVPN management interface.',
        env_var='AAD_OVPN_SOCKET_PATH',
    )
    parser_openvpn.add_argument(
        '-p',
        '--ovpn-password',
        help='Password for OpenVPN management interface.',
        env_var='AAD_OVPN_PASSWORD',
    )

    parser_prometheus = parser.add_argument_group('Prometheus settings')
    parser_prometheus.add_argument(
        '--prometheus',
        action='store_true',
        env_var='AAD_PROMETHEUS_ENABLED',
        help='Enable prometheus statistics',
    )
    parser_prometheus.add_argument(
        '--prometheus-listen-addr',
        env_var='AAD_PROMETHEUS_LISTEN_HOST',
        default='',
        help='prometheus listen addr',
    )
    parser_prometheus.add_argument(
        '--prometheus-listen-port',
        type=int,
        env_var='AAD_PROMETHEUS_PORT',
        help=' prometheus statistics',
        default=9723,
    )
    parser_prometheus.add_argument(
        '--log-level',
        default=logging.INFO,
        type=lambda x: getattr(logging, x),
        env_var='AAD_LOG_LEVEL',
        help='Configure the logging level.',
    )

    options = parser.parse_args()

    # convert all configured loggers to use a background thread
    setup_logging_queues()

    logging.basicConfig(
        level=options.log_level, format='%(asctime)s %(levelname)s %(message)s'
    )

    if options.prometheus:
        start_http_server(
            options.prometheus_listen_port, options.prometheus_listen_addr
        )
        i = Info('openvpn_auth_azure_ad_version', 'info of openvpn-auth-azure-ad')
        i.info({'version': __version__})

    authenticator = TotpAuthenticator(
        mysql_host=options.mysql_host,
        mysql_username=options.mysql_username,
        mysql_password=options.mysql_password,
        mysql_database=options.mysql_database,
        threads=options.threads,
        host=options.ovpn_host,
        port=options.ovpn_port,
        unix_socket=options.ovpn_socket,
        password=options.ovpn_password,
    )
    if options.migrate:
        from pkg_resources import resource_string
        sql = resource_string('openvpn_auth_duo', 'schema.sql')
        sql = sql.decode('utf-8')
        authenticator.query(sql)
    elif options.enroll:
        email = options.email
        existing = authenticator.query(
            'select * from totp where email=%s',
            [email])
        if existing:
            secret_key = existing[0]['secret_key']
        else:
            secret_key = random_base32()
            authenticator.query(
                'insert into totp (email, secret_key)'
                ' values (%s, %s)', [email, secret_key])
        print(secret_key)
        totp = TOTP(secret_key)
        hostname = gethostname()
        uri = totp.provisioning_uri(
            name=email,
            issuer_name=f'openvpn {hostname}',
        )
        print(uri)
        qr_code(uri)
    elif options.reenroll:
        email = options.email
        existing = authenticator.query(
            'select * from totp where email=%s',
            [email])
        if existing:
            log.info('[enroll] %s exists', email)
            secret_key = random_base32()
            authenticator.query(
                'update totp set secret_key = %s'
                ' where email=%s', [secret_key, email])
            print(secret_key)
            totp = TOTP(secret_key)
            hostname = gethostname()
            uri = totp.provisioning_uri(
                name=email,
                issuer_name=f'openvpn {hostname}',
            )
            print(uri)
            qr_code(uri)
        else:
            log.info('[enroll] %s does not exist', email)
    else:
        authenticator.run()



# from: https://stackoverflow.com/a/24457608/8087167

import logging
from concurrent.futures import ThreadPoolExecutor


log = logging.getLogger(__name__)


class ThreadPoolExecutorStackTraced(ThreadPoolExecutor):
    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""

        return super().submit(self._function_wrapper, fn, *args, **kwargs)

    def _function_wrapper(self, fn, *args, **kwargs):
        """Wraps `fn` in order to preserve the traceback of any kind of
        raised exception

        """
        try:
            return fn(*args, **kwargs)
        except Exception as ex:
            log.exception('Exception in thread: %s', ex)
